use std::{collections::BTreeSet, convert::Infallible, fmt::Display, str::FromStr};

use serde::{Deserialize, Serialize};
use strum::EnumString;
use ts_rs::TS;
use utoipa::ToSchema;

#[derive(
    Debug, Clone, Copy, Serialize, Deserialize, PartialEq, Eq, Hash, PartialOrd, Ord, TS, ToSchema,
)]
pub enum Permission {
    Read(PermissionDetail),
    Write(PermissionDetail),
    None,
}

impl FromStr for Permission {
    type Err = Infallible;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        if let Some((permission, detail)) = s.split_once(':') {
            match permission {
                "Read" => Ok(Self::Read(
                    PermissionDetail::from_str(detail).unwrap_or_default(),
                )),
                "Write" => Ok(Self::Write(
                    PermissionDetail::from_str(detail).unwrap_or_default(),
                )),
                _ => Ok(Self::None),
            }
        } else {
            Ok(Self::None)
        }
    }
}

impl Display for Permission {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let string = match self {
            Permission::Read(detail) => format!("Read:{detail}"),
            Permission::Write(detail) => format!("Write:{detail}"),
            Permission::None => "None".to_string(),
        };
        f.write_str(&string)
    }
}

#[derive(
    Debug,
    Default,
    Clone,
    Copy,
    Serialize,
    Deserialize,
    PartialEq,
    Eq,
    PartialOrd,
    Ord,
    Hash,
    EnumString,
    strum::Display,
    TS,
    ToSchema,
)]
pub enum PermissionDetail {
    Users,
    APIKeys,
    Signature,
    Departments,
    Units,
    Vendors,
    #[default]
    None,
}

#[derive(Debug, Clone, Serialize, Deserialize, TS, ToSchema)]
pub struct PermissionContainer(pub BTreeSet<Permission>);

impl From<Option<Vec<String>>> for PermissionContainer {
    fn from(value: Option<Vec<String>>) -> Self {
        let set = match value {
            Some(values) => BTreeSet::from_iter(
                values
                    .iter()
                    .map(|s| Permission::from_str(s).unwrap_or(Permission::None)),
            ),
            None => BTreeSet::new(),
        };

        Self(set)
    }
}

#[cfg(test)]
mod test {
    use std::{
        collections::{BTreeSet, HashSet},
        str::FromStr,
    };

    use chrono::Local;

    use crate::api::{
        backend::integrity_verification::IntegrityVerifier,
        routes::{
            departments::models::ShortDepartment,
            models::Status,
            users::{
                models::{DepartmentContainer, GroupContainer, User},
                permissions::{PermissionContainer, PermissionDetail},
            },
        },
    };

    use super::Permission;

    #[tokio::test]
    async fn test_permissions() {
        let permission = Permission::Read(PermissionDetail::Users);
        assert_eq!("Read:Users", permission.to_string());

        let parsed = Permission::from_str("Write:Users").unwrap();
        assert_eq!(parsed, Permission::Write(PermissionDetail::Users));
    }

    #[tokio::test]
    async fn test_integrity_validation() {
        let mut user = User {
            user_id: "id".to_string(),
            active_directory_auth: false,
            password: "Some PW".to_string(),
            name: "Clark".to_string(),
            surname: "Kent".to_string(),
            email: "Superman@heros.com".to_string(),
            department: DepartmentContainer(Some(ShortDepartment {
                id: 1,
                name: "Some name".to_string(),
                description: None,
            })),
            groups: GroupContainer(HashSet::new()),
            group_permissions: PermissionContainer(BTreeSet::new()),
            permissions: PermissionContainer(BTreeSet::from_iter([Permission::Read(
                PermissionDetail::Departments,
            )])),
            status_flag: Status::Active,
            creation_date: Local::now().naive_local(),
            last_change: Local::now().naive_local(),
        };

        let verifier = IntegrityVerifier::new();

        let hash = verifier.sign(&user);

        assert!(verifier.verify(&user, hash).is_ok());

        user.permissions
            .0
            .insert(Permission::Write(PermissionDetail::Units));

        assert!(verifier.verify(&user, hash).is_err());
    }
}
